"""Test the interactive interpreter."""

import sys
import os
import unittest
import subprocess
from textwrap import dedent
from test.support import cpython_only, SuppressCrashReport
from test.support.script_helper import kill_python

def spawn_repl(*args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kw):
    """Run the Python REPL with the given arguments.

    kw is extra keyword args to pass to subprocess.Popen. Returns a Popen
    object.
    """

    # To run the REPL without using a terminal, spawn python with the command
    # line option '-i' and the process name set to '<stdin>'.
    # The directory of argv[0] must match the directory of the Python
    # executable for the Popen() call to python to succeed as the directory
    # path may be used by Py_GetPath() to build the default module search
    # path.
    stdin_fname = os.path.join(os.path.dirname(sys.executable), "<stdin>")
    cmd_line = [stdin_fname, '-E', '-i']
    cmd_line.extend(args)

    # Set TERM=vt100, for the rationale see the comments in spawn_python() of
    # test.support.script_helper.
    env = kw.setdefault('env', dict(os.environ))
    env['TERM'] = 'vt100'
    return subprocess.Popen(cmd_line, executable=sys.executable,
                            stdin=subprocess.PIPE,
                            stdout=stdout, stderr=stderr,
                            **kw)

class TestInteractiveInterpreter(unittest.TestCase):

    @cpython_only
    def test_no_memory(self):
        # Issue #30696: Fix the interactive interpreter looping endlessly when
        # no memory. Check also that the fix does not break the interactive
        # loop when an exception is raised.
        user_input = """
            import sys, _testcapi
            1/0
            print('After the exception.')
            _testcapi.set_nomemory(0)
            sys.exit(0)
        """
        user_input = dedent(user_input)
        user_input = user_input.encode()
        p = spawn_repl()
        with SuppressCrashReport():
            p.stdin.write(user_input)
        output = kill_python(p)
        self.assertIn(b'After the exception.', output)
        # Exit code 120: Py_FinalizeEx() failed to flush stdout and stderr.
        self.assertIn(p.returncode, (1, 120))

if __name__ == "__main__":
    unittest.main()



import faulthandler
import importlib
import io
import os
import sys
import time
import traceback
import unittest
from test import support
from test.libregrtest.refleak import dash_R, clear_caches
from test.libregrtest.save_env import saved_test_environment


# Test result constants.
PASSED = 1
FAILED = 0
ENV_CHANGED = -1
SKIPPED = -2
RESOURCE_DENIED = -3
INTERRUPTED = -4
CHILD_ERROR = -5   # error in a child process

_FORMAT_TEST_RESULT = {
    PASSED: '%s passed',
    FAILED: '%s failed',
    ENV_CHANGED: '%s failed (env changed)',
    SKIPPED: '%s skipped',
    RESOURCE_DENIED: '%s skipped (resource denied)',
    INTERRUPTED: '%s interrupted',
    CHILD_ERROR: '%s crashed',
}

# Minimum duration of a test to display its duration or to mention that
# the test is running in background
PROGRESS_MIN_TIME = 30.0   # seconds

# small set of tests to determine if we have a basically functioning interpreter
# (i.e. if any of these fail, then anything else is likely to follow)
STDTESTS = [
    'test_grammar',
    'test_opcodes',
    'test_dict',
    'test_builtin',
    'test_exceptions',
    'test_types',
    'test_unittest',
    'test_doctest',
    'test_doctest2',
    'test_support'
]

# set of tests that we don't want to be executed when using regrtest
NOTTESTS = set()


def format_test_result(test_name, result):
    fmt = _FORMAT_TEST_RESULT.get(result, "%s")
    return fmt % test_name


def findtests(testdir=None, stdtests=STDTESTS, nottests=NOTTESTS):
    """Return a list of all applicable test modules."""
    testdir = findtestdir(testdir)
    names = os.listdir(testdir)
    tests = []
    others = set(stdtests) | nottests
    for name in names:
        mod, ext = os.path.splitext(name)
        if mod[:5] == "test_" and ext in (".py", "") and mod not in others:
            tests.append(mod)
    return stdtests + sorted(tests)


def get_abs_module(ns, test):
    if test.startswith('test.') or ns.testdir:
        return test
    else:
        # Always import it from the test package
        return 'test.' + test


def runtest(ns, test):
    """Run a single test.

    ns -- regrtest namespace of options
    test -- the name of the test

    Returns the tuple (result, test_time, xml_data), where result is one
    of the constants:

        INTERRUPTED      KeyboardInterrupt when run under -j
        RESOURCE_DENIED  test skipped because resource denied
        SKIPPED          test skipped for some other reason
        ENV_CHANGED      test failed because it changed the execution environment
        FAILED           test failed
        PASSED           test passed

    If ns.xmlpath is not None, xml_data is a list containing each
    generated testsuite element.
    """

    output_on_failure = ns.verbose3

    use_timeout = (ns.timeout is not None)
    if use_timeout:
        faulthandler.dump_traceback_later(ns.timeout, exit=True)
    try:
        support.set_match_tests(ns.match_tests)
        # reset the environment_altered flag to detect if a test altered
        # the environment
        support.environment_altered = False
        support.junit_xml_list = xml_list = [] if ns.xmlpath else None
        if ns.failfast:
            support.failfast = True
        if output_on_failure:
            support.verbose = True

            stream = io.StringIO()
            orig_stdout = sys.stdout
            orig_stderr = sys.stderr
            try:
                sys.stdout = stream
                sys.stderr = stream
                result = runtest_inner(ns, test, display_failure=False)
                if result[0] != PASSED:
                    output = stream.getvalue()
                    orig_stderr.write(output)
                    orig_stderr.flush()
            finally:
                sys.stdout = orig_stdout
                sys.stderr = orig_stderr
        else:
            support.verbose = ns.verbose  # Tell tests to be moderately quiet
            result = runtest_inner(ns, test, display_failure=not ns.verbose)

        if xml_list:
            import xml.etree.ElementTree as ET
            xml_data = [ET.tostring(x).decode('us-ascii') for x in xml_list]
        else:
            xml_data = None
        return result + (xml_data,)
    finally:
        if use_timeout:
            faulthandler.cancel_dump_traceback_later()
        cleanup_test_droppings(test, ns.verbose)
        support.junit_xml_list = None


def post_test_cleanup():
    support.reap_children()


def runtest_inner(ns, test, display_failure=True):
    support.unload(test)

    test_time = 0.0
    refleak = False  # True if the test leaked references.
    try:
        abstest = get_abs_module(ns, test)
        clear_caches()
        with saved_test_environment(test, ns.verbose, ns.quiet, pgo=ns.pgo) as environment:
            start_time = time.time()
            the_module = importlib.import_module(abstest)
            # If the test has a test_main, that will run the appropriate
            # tests.  If not, use normal unittest test loading.
            test_runner = getattr(the_module, "test_main", None)
            if test_runner is None:
                def test_runner():
                    loader = unittest.TestLoader()
                    tests = loader.loadTestsFromModule(the_module)
                    for error in loader.errors:
                        print(error, file=sys.stderr)
                    if loader.errors:
                        raise Exception("errors while loading tests")
                    support.run_unittest(tests)
            if ns.huntrleaks:
                refleak = dash_R(the_module, test, test_runner, ns.huntrleaks)
            else:
                test_runner()
            test_time = time.time() - start_time
        post_test_cleanup()
    except support.ResourceDenied as msg:
        if not ns.quiet and not ns.pgo:
            print(test, "skipped --", msg, flush=True)
        return RESOURCE_DENIED, test_time
    except unittest.SkipTest as msg:
        if not ns.quiet and not ns.pgo:
            print(test, "skipped --", msg, flush=True)
        return SKIPPED, test_time
    except KeyboardInterrupt:
        raise
    except support.TestFailed as msg:
        if not ns.pgo:
            if display_failure:
                print("test", test, "failed --", msg, file=sys.stderr,
                      flush=True)
            else:
                print("test", test, "failed", file=sys.stderr, flush=True)
        return FAILED, test_time
    except:
        msg = traceback.format_exc()
        if not ns.pgo:
            print("test", test, "crashed --", msg, file=sys.stderr,
                  flush=True)
        return FAILED, test_time
    else:
        if refleak:
            return FAILED, test_time
        if environment.changed:
            return ENV_CHANGED, test_time
        return PASSED, test_time


def cleanup_test_droppings(testname, verbose):
    import shutil
    import stat
    import gc

    # First kill any dangling references to open files etc.
    # This can also issue some ResourceWarnings which would otherwise get
    # triggered during the following test run, and possibly produce failures.
    gc.collect()

    # Try to clean up junk commonly left behind.  While tests shouldn't leave
    # any files or directories behind, when a test fails that can be tedious
    # for it to arrange.  The consequences can be especially nasty on Windows,
    # since if a test leaves a file open, it cannot be deleted by name (while
    # there's nothing we can do about that here either, we can display the
    # name of the offending test, which is a real help).
    for name in (support.TESTFN,
                 "db_home",
                ):
        if not os.path.exists(name):
            continue

        if os.path.isdir(name):
            kind, nuker = "directory", shutil.rmtree
        elif os.path.isfile(name):
            kind, nuker = "file", os.unlink
        else:
            raise SystemError("os.path says %r exists but is neither "
                              "directory nor file" % name)

        if verbose:
            print("%r left behind %s %r" % (testname, kind, name))
        try:
            # if we have chmod, fix possible permissions problems
            # that might prevent cleanup
            if (hasattr(os, 'chmod')):
                os.chmod(name, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
            nuker(name)
        except Exception as msg:
            print(("%r left behind %s %r and it couldn't be "
                "removed: %s" % (testname, kind, name, msg)), file=sys.stderr)


def findtestdir(path=None):
    return path or os.path.dirname(os.path.dirname(__file__)) or os.curdir




## this python script can be used with icsClientPython_3
##      to test a connection to an scpi module connected at ttyS5

# The connection parameters must be given when the example is launched example:
#       python3 001-iCS.py -c ws://192.168.0.25:8080
#


import sys
import time
import math
import argparse
import traceback
import importlib
import types
import os

import clientCommon

connectionName = "new_iCS_connection"
#connectionParameters = "ws://ics.iseg-hv.com:8080"
connectionParameters = "ws://localhost:1234"
connectionLogin = "demo"
connectionPassword = "demo"

for argvIndex in range(0, len(sys.argv)):
    if argvIndex + 1 < len(sys.argv):
        if sys.argv[argvIndex] == "-c":
            connectionParameters = sys.argv[argvIndex + 1]

## this must specify the location of the icsClientPython_3 python module
pythonModulePath = "../platform/linux/64/"
if clientCommon.checkIsWindowsPlatform():
    pythonModulePath = "../bin/"
pythonModulePath = os.path.abspath(pythonModulePath)

## import icsClient module 
icsClient = clientCommon.importFromPath(pythonModulePath, "icsClientPython_3")
if icsClient == 0:
    print("Failed to import icsClient module => Exit.")
    exit(1)

## establish a connection and wait until it is ready (with timeout)
connection = clientCommon.establishConnection(icsClient, connectionName, connectionParameters, connectionLogin, connectionPassword)
if type(connection) != icsClient.connection:
    print("Failed to connect => Exit.")
    exit(1)

## 
connection.rawCommand("{\"i\":\"<session_id>\",\"t\":\"request\",\"c\":[{\"c\":\"getItem\",\"p\":{\"p\":{\"l\":\"0\",\"a\":\"1000\",\"c\":\"\"},\"i\":\"Control.power\",\"v\":\"\",\"u\":\"\"}}],\"r\":\"websocket\"}")

all_modules = connection.getModules()
print("All modules {0}.".format(all_modules))

all_crates = connection.getCrates()
print("All crates {0}.".format(all_crates))

clientCommon.disconnectAll(icsClient)




# Run the _testcapi module tests (tests for the Python/C API):  by defn,
# these are all functions _testcapi exports whose name begins with 'test_'.

from collections import OrderedDict
import os
import pickle
import random
import re
import subprocess
import sys
import sysconfig
import textwrap
import threading
import time
import unittest
from test import support
from test.support import MISSING_C_DOCSTRINGS
from test.support.script_helper import assert_python_failure, assert_python_ok
try:
    import _posixsubprocess
except ImportError:
    _posixsubprocess = None

# Skip this test if the _testcapi module isn't available.
_testcapi = support.import_module('_testcapi')

# Were we compiled --with-pydebug or with #define Py_DEBUG?
Py_DEBUG = hasattr(sys, 'gettotalrefcount')


def testfunction(self):
    """some doc"""
    return self

class InstanceMethod:
    id = _testcapi.instancemethod(id)
    testfunction = _testcapi.instancemethod(testfunction)

class CAPITest(unittest.TestCase):

    def test_instancemethod(self):
        inst = InstanceMethod()
        self.assertEqual(id(inst), inst.id())
        self.assertTrue(inst.testfunction() is inst)
        self.assertEqual(inst.testfunction.__doc__, testfunction.__doc__)
        self.assertEqual(InstanceMethod.testfunction.__doc__, testfunction.__doc__)

        InstanceMethod.testfunction.attribute = "test"
        self.assertEqual(testfunction.attribute, "test")
        self.assertRaises(AttributeError, setattr, inst.testfunction, "attribute", "test")

    def test_no_FatalError_infinite_loop(self):
        with support.SuppressCrashReport():
            p = subprocess.Popen([sys.executable, "-c",
                                  'import _testcapi;'
                                  '_testcapi.crash_no_current_thread()'],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        self.assertEqual(out, b'')
        # This used to cause an infinite loop.
        self.assertTrue(err.rstrip().startswith(
                         b'Fatal Python error:'
                         b' PyThreadState_Get: no current thread'))

    def test_memoryview_from_NULL_pointer(self):
        self.assertRaises(ValueError, _testcapi.make_memoryview_from_NULL_pointer)

    def test_exc_info(self):
        raised_exception = ValueError("5")
        new_exc = TypeError("TEST")
        try:
            raise raised_exception
        except ValueError as e:
            tb = e.__traceback__
            orig_sys_exc_info = sys.exc_info()
            orig_exc_info = _testcapi.set_exc_info(new_exc.__class__, new_exc, None)
            new_sys_exc_info = sys.exc_info()
            new_exc_info = _testcapi.set_exc_info(*orig_exc_info)
            reset_sys_exc_info = sys.exc_info()

            self.assertEqual(orig_exc_info[1], e)

            self.assertSequenceEqual(orig_exc_info, (raised_exception.__class__, raised_exception, tb))
            self.assertSequenceEqual(orig_sys_exc_info, orig_exc_info)
            self.assertSequenceEqual(reset_sys_exc_info, orig_exc_info)
            self.assertSequenceEqual(new_exc_info, (new_exc.__class__, new_exc, None))
            self.assertSequenceEqual(new_sys_exc_info, new_exc_info)
        else:
            self.assertTrue(False)

    @unittest.skipUnless(_posixsubprocess, '_posixsubprocess required for this test.')
    def test_seq_bytes_to_charp_array(self):
        # Issue #15732: crash in _PySequence_BytesToCharpArray()
        class Z(object):
            def __len__(self):
                return 1
        self.assertRaises(TypeError, _posixsubprocess.fork_exec,
                          1,Z(),3,(1, 2),5,6,7,8,9,10,11,12,13,14,15,16,17)
        # Issue #15736: overflow in _PySequence_BytesToCharpArray()
        class Z(object):
            def __len__(self):
                return sys.maxsize
            def __getitem__(self, i):
                return b'x'
        self.assertRaises(MemoryError, _posixsubprocess.fork_exec,
                          1,Z(),3,(1, 2),5,6,7,8,9,10,11,12,13,14,15,16,17)

    @unittest.skipUnless(_posixsubprocess, '_posixsubprocess required for this test.')
    def test_subprocess_fork_exec(self):
        class Z(object):
            def __len__(self):
                return 1

        # Issue #15738: crash in subprocess_fork_exec()
        self.assertRaises(TypeError, _posixsubprocess.fork_exec,
                          Z(),[b'1'],3,(1, 2),5,6,7,8,9,10,11,12,13,14,15,16,17)

    @unittest.skipIf(MISSING_C_DOCSTRINGS,
                     "Signature information for builtins requires docstrings")
    def test_docstring_signature_parsing(self):

        self.assertEqual(_testcapi.no_docstring.__doc__, None)
        self.assertEqual(_testcapi.no_docstring.__text_signature__, None)

        self.assertEqual(_testcapi.docstring_empty.__doc__, None)
        self.assertEqual(_testcapi.docstring_empty.__text_signature__, None)

        self.assertEqual(_testcapi.docstring_no_signature.__doc__,
            "This docstring has no signature.")
        self.assertEqual(_testcapi.docstring_no_signature.__text_signature__, None)

        self.assertEqual(_testcapi.docstring_with_invalid_signature.__doc__,
            "docstring_with_invalid_signature($module, /, boo)\n"
            "\n"
            "This docstring has an invalid signature."
            )
        self.assertEqual(_testcapi.docstring_with_invalid_signature.__text_signature__, None)

        self.assertEqual(_testcapi.docstring_with_invalid_signature2.__doc__,
            "docstring_with_invalid_signature2($module, /, boo)\n"
            "\n"
            "--\n"
            "\n"
            "This docstring also has an invalid signature."
            )
        self.assertEqual(_testcapi.docstring_with_invalid_signature2.__text_signature__, None)

        self.assertEqual(_testcapi.docstring_with_signature.__doc__,
            "This docstring has a valid signature.")
        self.assertEqual(_testcapi.docstring_with_signature.__text_signature__, "($module, /, sig)")

        self.assertEqual(_testcapi.docstring_with_signature_but_no_doc.__doc__, None)
        self.assertEqual(_testcapi.docstring_with_signature_but_no_doc.__text_signature__,
            "($module, /, sig)")

        self.assertEqual(_testcapi.docstring_with_signature_and_extra_newlines.__doc__,
            "\nThis docstring has a valid signature and some extra newlines.")
        self.assertEqual(_testcapi.docstring_with_signature_and_extra_newlines.__text_signature__,
            "($module, /, parameter)")

    def test_c_type_with_matrix_multiplication(self):
        M = _testcapi.matmulType
        m1 = M()
        m2 = M()
        self.assertEqual(m1 @ m2, ("matmul", m1, m2))
        self.assertEqual(m1 @ 42, ("matmul", m1, 42))
        self.assertEqual(42 @ m1, ("matmul", 42, m1))
        o = m1
        o @= m2
        self.assertEqual(o, ("imatmul", m1, m2))
        o = m1
        o @= 42
        self.assertEqual(o, ("imatmul", m1, 42))
        o = 42
        o @= m1
        self.assertEqual(o, ("matmul", 42, m1))

    def test_return_null_without_error(self):
        # Issue #23571: A function must not return NULL without setting an
        # error
        if Py_DEBUG:
            code = textwrap.dedent("""
                import _testcapi
                from test import support

                with support.SuppressCrashReport():
                    _testcapi.return_null_without_error()
            """)
            rc, out, err = assert_python_failure('-c', code)
            self.assertRegex(err.replace(b'\r', b''),
                             br'Fatal Python error: a function returned NULL '
                                br'without setting an error\n'
                             br'SystemError: <built-in function '
                                 br'return_null_without_error> returned NULL '
                                 br'without setting an error\n'
                             br'\n'
                             br'Current thread.*:\n'
                             br'  File .*", line 6 in <module>')
        else:
            with self.assertRaises(SystemError) as cm:
                _testcapi.return_null_without_error()
            self.assertRegex(str(cm.exception),
                             'return_null_without_error.* '
                             'returned NULL without setting an error')

    def test_return_result_with_error(self):
        # Issue #23571: A function must not return a result with an error set
        if Py_DEBUG:
            code = textwrap.dedent("""
                import _testcapi
                from test import support

                with support.SuppressCrashReport():
                    _testcapi.return_result_with_error()
            """)
            rc, out, err = assert_python_failure('-c', code)
            self.assertRegex(err.replace(b'\r', b''),
                             br'Fatal Python error: a function returned a '
                                br'result with an error set\n'
                             br'ValueError\n'
                             br'\n'
                             br'The above exception was the direct cause '
                                br'of the following exception:\n'
                             br'\n'
                             br'SystemError: <built-in '
                                br'function return_result_with_error> '
                                br'returned a result with an error set\n'
                             br'\n'
                             br'Current thread.*:\n'
                             br'  File .*, line 6 in <module>')
        else:
            with self.assertRaises(SystemError) as cm:
                _testcapi.return_result_with_error()
            self.assertRegex(str(cm.exception),
                             'return_result_with_error.* '
                             'returned a result with an error set')

    def test_buildvalue_N(self):
        _testcapi.test_buildvalue_N()

    def test_set_nomemory(self):
        code = """if 1:
            import _testcapi

            class C(): pass

            # The first loop tests both functions and that remove_mem_hooks()
            # can be called twice in a row. The second loop checks a call to
            # set_nomemory() after a call to remove_mem_hooks(). The third
            # loop checks the start and stop arguments of set_nomemory().
            for outer_cnt in range(1, 4):
                start = 10 * outer_cnt
                for j in range(100):
                    if j == 0:
                        if outer_cnt != 3:
                            _testcapi.set_nomemory(start)
                        else:
                            _testcapi.set_nomemory(start, start + 1)
                    try:
                        C()
                    except MemoryError as e:
                        if outer_cnt != 3:
                            _testcapi.remove_mem_hooks()
                        print('MemoryError', outer_cnt, j)
                        _testcapi.remove_mem_hooks()
                        break
        """
        rc, out, err = assert_python_ok('-c', code)
        self.assertIn(b'MemoryError 1 10', out)
        self.assertIn(b'MemoryError 2 20', out)
        self.assertIn(b'MemoryError 3 30', out)

    def test_mapping_keys_values_items(self):
        class Mapping1(dict):
            def keys(self):
                return list(super().keys())
            def values(self):
                return list(super().values())
            def items(self):
                return list(super().items())
        class Mapping2(dict):
            def keys(self):
                return tuple(super().keys())
            def values(self):
                return tuple(super().values())
            def items(self):
                return tuple(super().items())
        dict_obj = {'foo': 1, 'bar': 2, 'spam': 3}

        for mapping in [{}, OrderedDict(), Mapping1(), Mapping2(),
                        dict_obj, OrderedDict(dict_obj),
                        Mapping1(dict_obj), Mapping2(dict_obj)]:
            self.assertListEqual(_testcapi.get_mapping_keys(mapping),
                                 list(mapping.keys()))
            self.assertListEqual(_testcapi.get_mapping_values(mapping),
                                 list(mapping.values()))
            self.assertListEqual(_testcapi.get_mapping_items(mapping),
                                 list(mapping.items()))

    def test_mapping_keys_values_items_bad_arg(self):
        self.assertRaises(AttributeError, _testcapi.get_mapping_keys, None)
        self.assertRaises(AttributeError, _testcapi.get_mapping_values, None)
        self.assertRaises(AttributeError, _testcapi.get_mapping_items, None)

        class BadMapping:
            def keys(self):
                return None
            def values(self):
                return None
            def items(self):
                return None
        bad_mapping = BadMapping()
        self.assertRaises(TypeError, _testcapi.get_mapping_keys, bad_mapping)
        self.assertRaises(TypeError, _testcapi.get_mapping_values, bad_mapping)
        self.assertRaises(TypeError, _testcapi.get_mapping_items, bad_mapping)


class TestPendingCalls(unittest.TestCase):

    def pendingcalls_submit(self, l, n):
        def callback():
            #this function can be interrupted by thread switching so let's
            #use an atomic operation
            l.append(None)

        for i in range(n):
            time.sleep(random.random()*0.02) #0.01 secs on average
            #try submitting callback until successful.
            #rely on regular interrupt to flush queue if we are
            #unsuccessful.
            while True:
                if _testcapi._pending_threadfunc(callback):
                    break;

    def pendingcalls_wait(self, l, n, context = None):
        #now, stick around until l[0] has grown to 10
        count = 0;
        while len(l) != n:
            #this busy loop is where we expect to be interrupted to
            #run our callbacks.  Note that callbacks are only run on the
            #main thread
            if False and support.verbose:
                print("(%i)"%(len(l),),)
            for i in range(1000):
                a = i*i
            if context and not context.event.is_set():
                continue
            count += 1
            self.assertTrue(count < 10000,
                "timeout waiting for %i callbacks, got %i"%(n, len(l)))
        if False and support.verbose:
            print("(%i)"%(len(l),))

    def test_pendingcalls_threaded(self):

        #do every callback on a separate thread
        n = 32 #total callbacks
        threads = []
        class foo(object):pass
        context = foo()
        context.l = []
        context.n = 2 #submits per thread
        context.nThreads = n // context.n
        context.nFinished = 0
        context.lock = threading.Lock()
        context.event = threading.Event()

        threads = [threading.Thread(target=self.pendingcalls_thread,
                                    args=(context,))
                   for i in range(context.nThreads)]
        with support.start_threads(threads):
            self.pendingcalls_wait(context.l, n, context)

    def pendingcalls_thread(self, context):
        try:
            self.pendingcalls_submit(context.l, context.n)
        finally:
            with context.lock:
                context.nFinished += 1
                nFinished = context.nFinished
                if False and support.verbose:
                    print("finished threads: ", nFinished)
            if nFinished == context.nThreads:
                context.event.set()

    def test_pendingcalls_non_threaded(self):
        #again, just using the main thread, likely they will all be dispatched at
        #once.  It is ok to ask for too many, because we loop until we find a slot.
        #the loop can be interrupted to dispatch.
        #there are only 32 dispatch slots, so we go for twice that!
        l = []
        n = 64
        self.pendingcalls_submit(l, n)
        self.pendingcalls_wait(l, n)


class SubinterpreterTest(unittest.TestCase):

    def test_subinterps(self):
        import builtins
        r, w = os.pipe()
        code = """if 1:
            import sys, builtins, pickle
            with open({:d}, "wb") as f:
                pickle.dump(id(sys.modules), f)
                pickle.dump(id(builtins), f)
            """.format(w)
        with open(r, "rb") as f:
            ret = support.run_in_subinterp(code)
            self.assertEqual(ret, 0)
            self.assertNotEqual(pickle.load(f), id(sys.modules))
            self.assertNotEqual(pickle.load(f), id(builtins))


class TestThreadState(unittest.TestCase):

    @support.reap_threads
    def test_thread_state(self):
        # some extra thread-state tests driven via _testcapi
        def target():
            idents = []

            def callback():
                idents.append(threading.get_ident())

            _testcapi._test_thread_state(callback)
            a = b = callback
            time.sleep(1)
            # Check our main thread is in the list exactly 3 times.
            self.assertEqual(idents.count(threading.get_ident()), 3,
                             "Couldn't find main thread correctly in the list")

        target()
        t = threading.Thread(target=target)
        t.start()
        t.join()


class Test_testcapi(unittest.TestCase):
    locals().update((name, getattr(_testcapi, name))
                    for name in dir(_testcapi)
                    if name.startswith('test_') and not name.endswith('_code'))


class PyMemDebugTests(unittest.TestCase):
    PYTHONMALLOC = 'debug'
    # '0x04c06e0' or '04C06E0'
    PTR_REGEX = r'(?:0x)?[0-9a-fA-F]+'

    def check(self, code):
        with support.SuppressCrashReport():
            out = assert_python_failure('-c', code,
                                        PYTHONMALLOC=self.PYTHONMALLOC)
        stderr = out.err
        return stderr.decode('ascii', 'replace')

    def test_buffer_overflow(self):
        out = self.check('import _testcapi; _testcapi.pymem_buffer_overflow()')
        regex = (r"Debug memory block at address p={ptr}: API 'm'\n"
                 r"    16 bytes originally requested\n"
                 r"    The [0-9] pad bytes at p-[0-9] are FORBIDDENBYTE, as expected.\n"
                 r"    The [0-9] pad bytes at tail={ptr} are not all FORBIDDENBYTE \(0x[0-9a-f]{{2}}\):\n"
                 r"        at tail\+0: 0x78 \*\*\* OUCH\n"
                 r"        at tail\+1: 0xfb\n"
                 r"        at tail\+2: 0xfb\n"
                 r"        .*\n"
                 r"    The block was made by call #[0-9]+ to debug malloc/realloc.\n"
                 r"    Data at p: cb cb cb .*\n"
                 r"\n"
                 r"Fatal Python error: bad trailing pad byte")
        regex = regex.format(ptr=self.PTR_REGEX)
        regex = re.compile(regex, flags=re.DOTALL)
        self.assertRegex(out, regex)

    def test_api_misuse(self):
        out = self.check('import _testcapi; _testcapi.pymem_api_misuse()')
        regex = (r"Debug memory block at address p={ptr}: API 'm'\n"
                 r"    16 bytes originally requested\n"
                 r"    The [0-9] pad bytes at p-[0-9] are FORBIDDENBYTE, as expected.\n"
                 r"    The [0-9] pad bytes at tail={ptr} are FORBIDDENBYTE, as expected.\n"
                 r"    The block was made by call #[0-9]+ to debug malloc/realloc.\n"
                 r"    Data at p: cb cb cb .*\n"
                 r"\n"
                 r"Fatal Python error: bad ID: Allocated using API 'm', verified using API 'r'\n")
        regex = regex.format(ptr=self.PTR_REGEX)
        self.assertRegex(out, regex)

    def check_malloc_without_gil(self, code):
        out = self.check(code)
        expected = ('Fatal Python error: Python memory allocator called '
                    'without holding the GIL')
        self.assertIn(expected, out)

    def test_pymem_malloc_without_gil(self):
        # Debug hooks must raise an error if PyMem_Malloc() is called
        # without holding the GIL
        code = 'import _testcapi; _testcapi.pymem_malloc_without_gil()'
        self.check_malloc_without_gil(code)

    def test_pyobject_malloc_without_gil(self):
        # Debug hooks must raise an error if PyObject_Malloc() is called
        # without holding the GIL
        code = 'import _testcapi; _testcapi.pyobject_malloc_without_gil()'
        self.check_malloc_without_gil(code)


class PyMemMallocDebugTests(PyMemDebugTests):
    PYTHONMALLOC = 'malloc_debug'


@unittest.skipUnless(support.with_pymalloc(), 'need pymalloc')
class PyMemPymallocDebugTests(PyMemDebugTests):
    PYTHONMALLOC = 'pymalloc_debug'


@unittest.skipUnless(Py_DEBUG, 'need Py_DEBUG')
class PyMemDefaultTests(PyMemDebugTests):
    # test default allocator of Python compiled in debug mode
    PYTHONMALLOC = ''


if __name__ == "__main__":
    unittest.main()



# Run the tests in Programs/_testembed.c (tests for the CPython embedding APIs)
from test import support
import unittest

from collections import namedtuple
import os
import re
import subprocess
import sys


class EmbeddingTestsMixin:
    def setUp(self):
        here = os.path.abspath(__file__)
        basepath = os.path.dirname(os.path.dirname(os.path.dirname(here)))
        exename = "_testembed"
        if sys.platform.startswith("win"):
            ext = ("_d" if "_d" in sys.executable else "") + ".exe"
            exename += ext
            exepath = os.path.dirname(sys.executable)
        else:
            exepath = os.path.join(basepath, "Programs")
        self.test_exe = exe = os.path.join(exepath, exename)
        if not os.path.exists(exe):
            self.skipTest("%r doesn't exist" % exe)
        # This is needed otherwise we get a fatal error:
        # "Py_Initialize: Unable to get the locale encoding
        # LookupError: no codec search functions registered: can't find encoding"
        self.oldcwd = os.getcwd()
        os.chdir(basepath)

    def tearDown(self):
        os.chdir(self.oldcwd)

    def run_embedded_interpreter(self, *args, env=None):
        """Runs a test in the embedded interpreter"""
        cmd = [self.test_exe]
        cmd.extend(args)
        if env is not None and sys.platform == 'win32':
            # Windows requires at least the SYSTEMROOT environment variable to
            # start Python.
            env = env.copy()
            env['SYSTEMROOT'] = os.environ['SYSTEMROOT']

        p = subprocess.Popen(cmd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=True,
                             env=env)
        (out, err) = p.communicate()
        if p.returncode != 0 and support.verbose:
            print(f"--- {cmd} failed ---")
            print(f"stdout:\n{out}")
            print(f"stderr:\n{err}")
            print(f"------")

        self.assertEqual(p.returncode, 0,
                         "bad returncode %d, stderr is %r" %
                         (p.returncode, err))
        return out, err

    def run_repeated_init_and_subinterpreters(self):
        out, err = self.run_embedded_interpreter("repeated_init_and_subinterpreters")
        self.assertEqual(err, "")

        # The output from _testembed looks like this:
        # --- Pass 0 ---
        # interp 0 <0x1cf9330>, thread state <0x1cf9700>: id(modules) = 139650431942728
        # interp 1 <0x1d4f690>, thread state <0x1d35350>: id(modules) = 139650431165784
        # interp 2 <0x1d5a690>, thread state <0x1d99ed0>: id(modules) = 139650413140368
        # interp 3 <0x1d4f690>, thread state <0x1dc3340>: id(modules) = 139650412862200
        # interp 0 <0x1cf9330>, thread state <0x1cf9700>: id(modules) = 139650431942728
        # --- Pass 1 ---
        # ...

        interp_pat = (r"^interp (\d+) <(0x[\dA-F]+)>, "
                      r"thread state <(0x[\dA-F]+)>: "
                      r"id\(modules\) = ([\d]+)$")
        Interp = namedtuple("Interp", "id interp tstate modules")

        numloops = 0
        current_run = []
        for line in out.splitlines():
            if line == "--- Pass {} ---".format(numloops):
                self.assertEqual(len(current_run), 0)
                if support.verbose > 1:
                    print(line)
                numloops += 1
                continue

            self.assertLess(len(current_run), 5)
            match = re.match(interp_pat, line)
            if match is None:
                self.assertRegex(line, interp_pat)

            # Parse the line from the loop.  The first line is the main
            # interpreter and the 3 afterward are subinterpreters.
            interp = Interp(*match.groups())
            if support.verbose > 1:
                print(interp)
            self.assertTrue(interp.interp)
            self.assertTrue(interp.tstate)
            self.assertTrue(interp.modules)
            current_run.append(interp)

            # The last line in the loop should be the same as the first.
            if len(current_run) == 5:
                main = current_run[0]
                self.assertEqual(interp, main)
                yield current_run
                current_run = []


class EmbeddingTests(EmbeddingTestsMixin, unittest.TestCase):
    def test_subinterps_main(self):
        for run in self.run_repeated_init_and_subinterpreters():
            main = run[0]

            self.assertEqual(main.id, '0')

    def test_subinterps_different_ids(self):
        for run in self.run_repeated_init_and_subinterpreters():
            main, *subs, _ = run

            mainid = int(main.id)
            for i, sub in enumerate(subs):
                self.assertEqual(sub.id, str(mainid + i + 1))

    def test_subinterps_distinct_state(self):
        for run in self.run_repeated_init_and_subinterpreters():
            main, *subs, _ = run

            if '0x0' in main:
                # XXX Fix on Windows (and other platforms): something
                # is going on with the pointers in Programs/_testembed.c.
                # interp.interp is 0x0 and interp.modules is the same
                # between interpreters.
                raise unittest.SkipTest('platform prints pointers as 0x0')

            for sub in subs:
                # A new subinterpreter may have the same
                # PyInterpreterState pointer as a previous one if
                # the earlier one has already been destroyed.  So
                # we compare with the main interpreter.  The same
                # applies to tstate.
                self.assertNotEqual(sub.interp, main.interp)
                self.assertNotEqual(sub.tstate, main.tstate)
                self.assertNotEqual(sub.modules, main.modules)

    def test_forced_io_encoding(self):
        # Checks forced configuration of embedded interpreter IO streams
        env = dict(os.environ, PYTHONIOENCODING="utf-8:surrogateescape")
        out, err = self.run_embedded_interpreter("forced_io_encoding", env=env)
        if support.verbose > 1:
            print()
            print(out)
            print(err)
        expected_stream_encoding = "utf-8"
        expected_errors = "surrogateescape"
        expected_output = '\n'.join([
        "--- Use defaults ---",
        "Expected encoding: default",
        "Expected errors: default",
        "stdin: {in_encoding}:{errors}",
        "stdout: {out_encoding}:{errors}",
        "stderr: {out_encoding}:backslashreplace",
        "--- Set errors only ---",
        "Expected encoding: default",
        "Expected errors: ignore",
        "stdin: {in_encoding}:ignore",
        "stdout: {out_encoding}:ignore",
        "stderr: {out_encoding}:backslashreplace",
        "--- Set encoding only ---",
        "Expected encoding: latin-1",
        "Expected errors: default",
        "stdin: latin-1:{errors}",
        "stdout: latin-1:{errors}",
        "stderr: latin-1:backslashreplace",
        "--- Set encoding and errors ---",
        "Expected encoding: latin-1",
        "Expected errors: replace",
        "stdin: latin-1:replace",
        "stdout: latin-1:replace",
        "stderr: latin-1:backslashreplace"])
        expected_output = expected_output.format(
                                in_encoding=expected_stream_encoding,
                                out_encoding=expected_stream_encoding,
                                errors=expected_errors)
        # This is useful if we ever trip over odd platform behaviour
        self.maxDiff = None
        self.assertEqual(out.strip(), expected_output)

    def test_pre_initialization_api(self):
        """
        Checks some key parts of the C-API that need to work before the runtine
        is initialized (via Py_Initialize()).
        """
        env = dict(os.environ, PYTHONPATH=os.pathsep.join(sys.path))
        out, err = self.run_embedded_interpreter("pre_initialization_api", env=env)
        if sys.platform == "win32":
            expected_path = self.test_exe
        else:
            expected_path = os.path.join(os.getcwd(), "spam")
        expected_output = f"sys.executable: {expected_path}\n"
        self.assertIn(expected_output, out)
        self.assertEqual(err, '')

    def test_pre_initialization_sys_options(self):
        """
        Checks that sys.warnoptions and sys._xoptions can be set before the
        runtime is initialized (otherwise they won't be effective).
        """
        env = dict(os.environ, PYTHONPATH=os.pathsep.join(sys.path))
        out, err = self.run_embedded_interpreter(
                        "pre_initialization_sys_options", env=env)
        expected_output = (
            "sys.warnoptions: ['once', 'module', 'default']\n"
            "sys._xoptions: {'not_an_option': '1', 'also_not_an_option': '2'}\n"
            "warnings.filters[:3]: ['default', 'module', 'once']\n"
        )
        self.assertIn(expected_output, out)
        self.assertEqual(err, '')

    def test_bpo20891(self):
        """
        bpo-20891: Calling PyGILState_Ensure in a non-Python thread before
        calling PyEval_InitThreads() must not crash. PyGILState_Ensure() must
        call PyEval_InitThreads() for us in this case.
        """
        out, err = self.run_embedded_interpreter("bpo20891")
        self.assertEqual(out, '')
        self.assertEqual(err, '')

    def test_initialize_twice(self):
        """
        bpo-33932: Calling Py_Initialize() twice should do nothing (and not
        crash!).
        """
        out, err = self.run_embedded_interpreter("initialize_twice")
        self.assertEqual(out, '')
        self.assertEqual(err, '')

    def test_initialize_pymain(self):
        """
        bpo-34008: Calling Py_Main() after Py_Initialize() must not fail.
        """
        out, err = self.run_embedded_interpreter("initialize_pymain")
        self.assertEqual(out.rstrip(), "Py_Main() after Py_Initialize: sys.argv=['-c', 'arg2']")
        self.assertEqual(err, '')


class InitConfigTests(EmbeddingTestsMixin, unittest.TestCase):
    maxDiff = 4096
    DEFAULT_CONFIG = {
        'install_signal_handlers': 1,
        'Py_IgnoreEnvironmentFlag': 0,
        'use_hash_seed': 0,
        'hash_seed': 0,
        'allocator': '(null)',
        'dev_mode': 0,
        'faulthandler': 0,
        'tracemalloc': 0,
        'import_time': 0,
        'show_ref_count': 0,
        'show_alloc_count': 0,
        'dump_refs': 0,
        'malloc_stats': 0,
        'utf8_mode': 0,

        'coerce_c_locale': 0,
        'coerce_c_locale_warn': 0,

        'program_name': './_testembed',
        'argc': 0,
        'argv': '[]',
        'program': '(null)',

        'Py_IsolatedFlag': 0,
        'Py_NoSiteFlag': 0,
        'Py_BytesWarningFlag': 0,
        'Py_InspectFlag': 0,
        'Py_InteractiveFlag': 0,
        'Py_OptimizeFlag': 0,
        'Py_DebugFlag': 0,
        'Py_DontWriteBytecodeFlag': 0,
        'Py_VerboseFlag': 0,
        'Py_QuietFlag': 0,
        'Py_NoUserSiteDirectory': 0,
        'Py_UnbufferedStdioFlag': 0,

        '_disable_importlib': 0,
        'Py_FrozenFlag': 0,
    }

    def check_config(self, testname, expected):
        env = dict(os.environ)
        for key in list(env):
            if key.startswith('PYTHON'):
                del env[key]
        # Disable C locale coercion and UTF-8 mode to not depend
        # on the current locale
        env['PYTHONCOERCECLOCALE'] = '0'
        env['PYTHONUTF8'] = '0'
        out, err = self.run_embedded_interpreter(testname, env=env)
        # Ignore err

        expected = dict(self.DEFAULT_CONFIG, **expected)
        for key, value in expected.items():
            expected[key] = str(value)

        config = {}
        for line in out.splitlines():
            key, value = line.split(' = ', 1)
            config[key] = value
        self.assertEqual(config, expected)

    def test_init_default_config(self):
        self.check_config("init_default_config", {})

    def test_init_global_config(self):
        config = {
            'program_name': './globalvar',
            'Py_NoSiteFlag': 1,
            'Py_BytesWarningFlag': 1,
            'Py_InspectFlag': 1,
            'Py_InteractiveFlag': 1,
            'Py_OptimizeFlag': 2,
            'Py_DontWriteBytecodeFlag': 1,
            'Py_VerboseFlag': 1,
            'Py_QuietFlag': 1,
            'Py_UnbufferedStdioFlag': 1,
            'utf8_mode': 1,
            'Py_NoUserSiteDirectory': 1,
            'Py_FrozenFlag': 1,
        }
        self.check_config("init_global_config", config)

    def test_init_from_config(self):
        config = {
            'install_signal_handlers': 0,
            'use_hash_seed': 1,
            'hash_seed': 123,
            'allocator': 'malloc_debug',
            'tracemalloc': 2,
            'import_time': 1,
            'show_ref_count': 1,
            'show_alloc_count': 1,
            'malloc_stats': 1,

            'utf8_mode': 1,

            'program_name': './conf_program_name',
            'program': 'conf_program',

            'faulthandler': 1,
        }
        self.check_config("init_from_config", config)

    def test_init_env(self):
        config = {
            'use_hash_seed': 1,
            'hash_seed': 42,
            'allocator': 'malloc_debug',
            'tracemalloc': 2,
            'import_time': 1,
            'malloc_stats': 1,
            'utf8_mode': 1,
            'Py_InspectFlag': 1,
            'Py_OptimizeFlag': 2,
            'Py_DontWriteBytecodeFlag': 1,
            'Py_VerboseFlag': 1,
            'Py_UnbufferedStdioFlag': 1,
            'Py_NoUserSiteDirectory': 1,
            'faulthandler': 1,
            'dev_mode': 1,
        }
        self.check_config("init_env", config)

    def test_init_dev_mode(self):
        config = {
            'dev_mode': 1,
            'faulthandler': 1,
            'allocator': 'debug',
        }
        self.check_config("init_dev_mode", config)

    def test_init_isolated(self):
        config = {
            'Py_IsolatedFlag': 1,
            'Py_IgnoreEnvironmentFlag': 1,
            'Py_NoUserSiteDirectory': 1,
        }
        self.check_config("init_isolated", config)


if __name__ == "__main__":
    unittest.main()




#
# This example implements a data logger. Voltage and current measurements are written to a CSV file
#       in regular time intervals.
#
# The connection parameters must be given when the example is launched example:
#       python3 pythonDataLogger.py -c ws://192.168.0.25:8080
#
# It makes use of the icsClientPython_3 Python library.
#
# Some of the parameters below might have to be adjusted.
#

import sys
import os
import io
import time
import types
import traceback

import clientCommon

from datetime import date

class CsvDataLogger() :

    m_status = ""
    m_flagStopScript = False
    m_timeStampWriteToFile = 0.0
    m_iCSClient = 0
    m_iCSConnection = 0
    m_csvFile = 0

    ## the location of the icsClientPython_3 must be specified below
    m_pythonModulePath = ""

    ## adjust login and password to allow a connection to the iCSservice
    m_login = "demo"
    m_password = "demo"

    ## defines how and where the results are saved
    ##      <creation_time_stamp> will be replaced
    m_fileSavePath = "data_log_<creation_time_stamp>.csv"

    ## th measured values will be written every 10 seconds
    m_timeIntervalWriteToFile = 10.0

    ## defines a file where information and errors are logged to
    m_logFilePath = "debug.txt"

    ## debug, standard or none
    m_logFileLevel = "debug"

    def __init__(self):
        self.m_status = ""

    def checkStatusConnectionFailed(self):
        if self.m_status == "connection failed":
            return True
        return False

    def checkLogFileDebugModeEnabled(self):
        if self.m_logFileLevel == "debug":
            return True
        return False

    def checkLogFileStandardModeEnabled(self):
        if self.m_logFileLevel == "standard":
            return True
        return False

    def logWithLevel(self, level, msg):
        if self.m_logFilePath != "":
            try :
                logFile = open(self.m_logFilePath, "a")
                logFile.write("# {0} {1} : {2}".format(level, time.asctime(), msg))
                logFile.write("\n")
                logFile.close()
            except Exception as exc:
                ## print traceback
                exc_type, exc_value, exc_traceback = sys.exc_info()
                lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
                print("exception caught in CsvDataLogger::logWithLevel()")
                for line in lines:
                    print(line.strip())

    def logInfo(self, msg):
        if self.checkLogFileDebugModeEnabled() or self.checkLogFileStandardModeEnabled():
            self.logWithLevel("info", msg)

    def logError(self, msg):
        if self.checkLogFileDebugModeEnabled() or self.checkLogFileStandardModeEnabled():
            self.logWithLevel("error", msg)

    def logDebug(self, msg):
        if self.checkLogFileDebugModeEnabled():
            self.logWithLevel("debug", msg)

    ## as soon as this script is launched a connection to the iCS is established
    ##      In addition to the standard connection parameters the session id "sessionIdSupervisor" that was passed as an argument to the script is also required.
    ##      This makes sure that we always know how the scrpit is supervised.
    def establishScriptConnection(self, connectionParameters, sessionIdSupervisor):
        self.logInfo("Script name == {0}".format(clientCommon.getScriptName()))
        if not isinstance(self.m_iCSClient, types.ModuleType):
            self.m_iCSClient = clientCommon.importFromPath(self.m_pythonModulePath, "icsClientPython_3")
            if not isinstance(self.m_iCSClient, types.ModuleType):
                self.m_errorMessage = "Failed to import icsClient module."
                self.logError(self.m_errorMessage)
        if isinstance(self.m_iCSClient, types.ModuleType):
            connectionName = "iCS_csv_logger"
            self.m_iCSConnection = clientCommon.establishScriptConnection(self.m_iCSClient, connectionName, connectionParameters, sessionIdSupervisor, self.m_login, self.m_password)
            if not isinstance(self.m_iCSConnection, self.m_iCSClient.connection):
                self.m_errorMessage = "Failed to connect to {0}.".format(connectionParameters)
                self.logError(self.m_errorMessage)
        if self.checkIsConnected():
            self.logInfo("Connection to {0} established with supervisor session ID {1}.".format(connectionParameters, sessionIdSupervisor))
            mySessionId = self.m_iCSConnection.getStatusSessionId()
            self.logInfo("mySessionId == {0}".format(mySessionId))
        else:
            self.logError("Failed to establish connection to {0} with supervisor session ID {1}, login \"{2}\" and password \"{3}\".".format(connectionParameters, sessionIdSupervisor, self.m_login, self.m_password))

    def checkIsConnected(self):
        if isinstance(self.m_iCSClient, types.ModuleType):
            if isinstance(self.m_iCSConnection, self.m_iCSClient.connection):
                return True
        return False

    def closeConnection(self):
        clientCommon.disconnectAll(self.m_iCSClient)
        self.m_iCSConnection = 0

    def sendRawCommand(self, json):
        if self.checkIsConnected():
            self.m_iCSConnection.rawCommand(json)
            return True
        return False

    def establishConnection(self, connectionParameters, sessionIdSupervisor):
        ## check if connection is already OK
        if self.checkIsConnected():
            return True
        ## previous attempt failed
        if self.checkStatusConnectionFailed():
            return False
        ## attempt to connect
        rc = self.establishScriptConnection(connectionParameters, sessionIdSupervisor)
        if self.checkIsConnected():
            self.m_status = "connection established"
            return True
        self.m_status = "connection failed"
        return False

    ## generates a message that will be sent to the script supervisor.
    def generateJsonStringWithPayload(self, payload):
        json = "{{\"i\":\"<session_id>\",\"t\":\"script\",\"c\":{{\"e\": \"scriptData\",\"d\":{0},\"n\":\"<script_file_name>\",\"i\":\"<session_id_script_supervisor>\"}}}}".format(payload)
        return json

    ## processes the messages received from iCS
    def processNextMessage(self, icsMsg):
        rc = False

        if isinstance(icsMsg, self.m_iCSClient.dataFrame):
            ## handle the message appropriately
            # self.logDebug("processing message {0}".format(msg.m_icsMsg))
            if icsMsg.item == "script.stop":
                self.m_flagStopScript = True
                rc = True
            if icsMsg.item == "script.status":
                payload = "{{\"s\":\"{0}\"}}".format(self.m_status)
                json = self.generateJsonStringWithPayload(payload)
                self.sendRawCommand(json)
                self.logDebug("Script status response == {0}.".format(json))
                return True

        return rc

    def getCurrentTimeStamp(self):
        ts = int(round(time.time() * 1000))
        floatTime = float(ts)
        floatTime = floatTime / 1000.0
        return floatTime

    def getColSeparator(self):
        return ";"

    ## get measured values and write them to csv file
    ##      in regular time intervals
    def writeCsvFile(self):
        ts = self.getCurrentTimeStamp()
        ## every 500 ms : get measurements, compute resistance, control hardware according to the current recipe
        if ts >= self.m_timeStampWriteToFile + self.m_timeIntervalWriteToFile:
            self.m_timeStampWriteToFile = ts
            if isinstance(self.m_csvFile, io.IOBase):
                self.m_csvFile.write("{0}".format(ts))
                allModules = self.m_iCSConnection.getModules()
                # print("allModules == {0}".format(allModules))
                for module in allModules:
                    for channel in module.channels:
                        self.m_csvFile.write(self.getColSeparator())
                        self.m_csvFile.write("{0}".format(channel.getStatusVoltageMeasure()))
                        self.m_csvFile.write(self.getColSeparator())
                        self.m_csvFile.write("{0}".format(channel.getStatusCurrentMeasure()))
                self.m_csvFile.write("\n")
                self.m_csvFile.flush()


    ## this is 1 step in the main loop of the Python script
    def singleStep(self):

        ## catch exceptions
        try :

            if self.checkIsConnected():
                ## get a message from the iCS (iCSservice or user interface)
                icsMsg = self.m_iCSConnection.getNextPassThroughMessage()

                ## process messages passed from the user interface of iCS
                ##      until the "script.stop" command is issued
                self.processNextMessage(icsMsg)

            if not self.m_flagStopScript:
                ## get measurements and write data to CSV file
                self.writeCsvFile()

                ## this makes sure ressources can be used by other processes.
                time.sleep(0.5)

        except Exception as exc:
            ## print traceback
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
            self.logError("software error : exception caught in CsvDataLogger::singleStep()")
            for line in lines:
                self.logDebug(line.strip())
            self.m_status = "software error"

    ## this is the main loop of the Python script
    ##
    ##      it runs until the "script.stop" command is issued
    def runMainLoop(self, connectionParameters, sessionIdSupervisor, configFileName):

        ## clear log file
        if self.checkLogFileDebugModeEnabled() or self.checkLogFileStandardModeEnabled():
            ## disable generation of log file if the directory doesn't exist
            if self.m_logFilePath != "":
                self.m_logFilePath = os.path.abspath(self.m_logFilePath)
                parentPath = os.path.dirname(self.m_logFilePath)
                if not os.path.exists(parentPath) :
                    print("ERROR : the parent directory of the log file file {0} doesn't exist.".format(self.m_logFilePath))
                    print("ERROR : writing to log file is disabled.")
                    self.m_logFilePath  = ""
            if self.m_logFilePath != "":
                ## always catch exceptions to make sure that the script doesn't close unexpectedly
                try :
                    logFile = open(self.m_logFilePath, "w")
                    logFile.write("# Log File created for iCSPythonDataLogger.py on {0}".format(time.asctime()))
                    logFile.write("\n")
                    logFile.close()
                except Exception as exc:
                    ## print traceback
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
                    print("exception caught in CsvDataLogger::runMainLoop()")
                    for line in lines:
                        print(line.strip())

        ## establish connection
        if self.establishConnection(connectionParameters, sessionIdSupervisor):

            ## open CSV file and create column header
            try :
                ## adjust the name of the created CSV file
                today = date.today()
                timeStampStr = today.strftime("%Y%m%d")
                self.m_fileSavePath = self.m_fileSavePath.replace("<creation_time_stamp>", timeStampStr);
                self.m_fileSavePath = os.path.abspath(self.m_fileSavePath)

                self.logInfo("writing CSV file {0}".format(self.m_fileSavePath))

                ## disable generation of csv file if the directory doesn't exist
                if self.m_fileSavePath != "":
                    parentPath = os.path.dirname(self.m_fileSavePath)
                    if not os.path.exists(parentPath) :
                        self.m_fileSavePath  = ""
                        self.logError("the parent directory of the CSV file {0} doesn't exist.".format(self.m_fileSavePath))
                        self.logError("writing to CSV file is disabled.")

                if self.m_fileSavePath != "":
                    self.m_csvFile = open(self.m_fileSavePath, "w")
                    self.m_csvFile.write("# CSV File created for iCSPythonDataLogger.py on {0}".format(time.asctime()))
                    self.m_csvFile.write("\n")

                    self.m_csvFile.write("\"time [s]\"")
                    allModules = self.m_iCSConnection.getModules()
                    # print("allModules == {0}".format(allModules))
                    for module in allModules:
                        for channel in module.channels:
                            stat = channel.getStatusRunningState()
                            self.m_csvFile.write(self.getColSeparator())
                            self.m_csvFile.write("\"Channel.{0}.Vmeas [V]\"".format(channel))
                            self.m_csvFile.write(self.getColSeparator())
                            self.m_csvFile.write("\"Channel.{0}.Imeas [A]\"".format(channel))
                    self.m_csvFile.write("\n")
                    self.m_csvFile.flush()

            except Exception as exc:
                ## print traceback
                exc_type, exc_value, exc_traceback = sys.exc_info()
                lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
                print("exception caught in CsvDataLogger::runMainLoop()")
                for line in lines:
                    print(line.strip())

            self.logInfo("Entering execution loop.")

            # now the main loop can be executed
            self.m_flagStopScript = False
            while not self.m_flagStopScript:
                self.singleStep()


            if isinstance(self.m_csvFile, io.IOBase):
                try :
                    self.m_csvFile.close()
                except Exception as exc:
                    print("exception caught in CsvDataLogger::runMainLoop()")


            self.logInfo("Execution loop left.")

        self.closeConnection()
        self.logInfo("Script stopped.")






# the connection parameters can be given on the command line
#connectionParameters = "ws://localhost:8080"
connectionParameters = "ws://ics.iseg.hv.com:8080"
# isn't required in normal operation
sessionIdSupervisor = ""
# isn't required in normal operation
configFileName = ""

for argvIndex in range(0, len(sys.argv)):
    # print("argument[{0}] == {1}".format(argvIndex, sys.argv[argvIndex]))
    if argvIndex + 1 < len(sys.argv):
        if sys.argv[argvIndex] == "-i":
            sessionIdSupervisor = sys.argv[argvIndex + 1]
        if sys.argv[argvIndex] == "-c":
            connectionParameters = sys.argv[argvIndex + 1]

app = CsvDataLogger()
## this must specify the location of the icsClientPython_3 python module
app.m_pythonModulePath = "../platform/linux/64/"
if clientCommon.checkIsWindowsPlatform():
    app.m_pythonModulePath = "../bin/"
app.m_pythonModulePath = os.path.abspath(app.m_pythonModulePath)
app.runMainLoop(connectionParameters, sessionIdSupervisor, configFileName)





import sys
import importlib
import time
import traceback
import os

def checkIsWindowsPlatform():
    if sys.platform.startswith('win'):
        return True
    elif sys.platform.startswith('linux'):
        return False
    elif sys.platform.startswith('darwin'):
        return False
    return None

def checkIsMacPlatform():
    if sys.platform.startswith('darwin'):
        return True
    elif sys.platform.startswith('linux'):
        return False
    elif sys.platform.startswith('win'):
        return False
    return None

def checkIsLinuxPlatform():
    if checkIsMacPlatform() :
         return False
    if checkIsWindowsPlatform() :
         return False
    return True

def checkIsArmPlatform():
    if os.uname()[4] == "armv7l"  :
        return True
    return False

def checkIs64BitPlatform():
    if sys.maxsize > 2**32 :
        return True
    return False

def getScriptName():
    script_path = sys.argv[0]
    script_path = os.path.abspath(script_path)
    script_name = os.path.basename(script_path)
    return script_name

def getSessionIdSupervisor():
    session_id_supervisor = ""
    for argvIndex in range(0, len(sys.argv)):
        if argvIndex + 1 < len(sys.argv):
            if sys.argv[argvIndex] == "-i":
                session_id_supervisor = sys.argv[argvIndex + 1]
    return session_id_supervisor

def getConnectionParameters():
    connection_parameters = ""
    for argvIndex in range(0, len(sys.argv)):
        if argvIndex + 1 < len(sys.argv):
            if sys.argv[argvIndex] == "-c":
                connection_parameters = sys.argv[argvIndex + 1]
    return connection_parameters

## this function imports the icsClient Python module
##      the path and module name are defined here if they aren't given above
def importFromPath(module_path, module_name):

    script_path = os.path.dirname(sys.argv[0])
    script_path = os.path.abspath(script_path)

    ## module_path defines the location of the python module
    if module_path == "" :
        if checkIsWindowsPlatform():
            module_path = 'D:\\work\\iseg-build-pro\\build32\\isegcontrol_DEV_171207_32\\bin\\'
        elif checkIsLinuxPlatform():
            module_path = os.path.dirname(script_path)
            module_path = os.path.join(module_path, "platform")
            module_path = os.path.join(module_path, "linux")
            if checkIs64BitPlatform():
                module_path = os.path.join(module_path, "64")
            else:
                module_path = os.path.join(module_path, "32")
        elif checkIsMacPlatform():
            module_path = '/Applications/isegcontrol.app/Contents/Frameworks/'

    if module_path != "" :
        sys.path.append(module_path)

    ## the module name (depends on the python version)
    if module_name == "" :
#        module_name = "icsClientPython_{0}".format(sys.version_info[0])
        module_name = "icsClientPython_{0}_for_iCS".format(sys.version_info[0])

    ## try to import the module
    try:
        icsClient =  importlib.import_module(module_name)

    except ImportError as exc:

        ## print traceback
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        print("".join('!! ' + line for line in lines))

        print("Failed to load the module \"{0}\"".format(module_name))
        print("   from path \"{0}\"".format(module_path))
        full_path = os.path.join(module_path, module_name)
        if checkIsLinuxPlatform():
            full_path = full_path + ".so"
        if not os.path.exists(full_path) :
            print("   {0} doesn't exist".format(full_path))
            print("   => make sure that the module is found at this location")
        print("= The Python modules links to shared Qt libraries")
        print("   => make sure the Qt libraries are found")
        if checkIsLinuxPlatform() :
            print("   => make sure the LD_LIBRARY_PATH environment variable is set properly.")
            try:
                path = os.environ['LD_LIBRARY_PATH']
            except KeyError as exc:
                path = "--undefined--"
            print("   => LD_LIBRARY_PATH == {0}".format(path))

        return 0

    print("Sucessfully imported module \"{0}\" (version {1}) from path \"{2}\"".format(module_name, icsClient.getVersion(), module_path))

    return icsClient


def establishConnection(icsClient, connection_name, connection_ip, connection_login, connection_password):

    # icsClient.connect(<connection name>, <ip:port>, <login>, < password>)
    rc = icsClient.connect(connection_name, connection_ip, connection_login, connection_password)
    if rc == 'error':
        print("The connection \"{0}\" already exists".format(connection_name))

    return waitForConnection(icsClient, connection_name, connection_ip)

def establishScriptConnection(icsClient, connection_name, connection_ip, session_id_client, login, password):

    # third connection is created when the script is launched from an iCSservice.
    # It allows to exchange any kind of data with the iCS.

    if connection_ip == "":
        connection_ip = "ws://localhost:8080"

    if connection_name == "":
        connection_name = "script_localhost"

    # icsClient.scriptConnect(<connection name>, <ip:port>, <script_name>)
    script_name = getScriptName();
    rc = icsClient.scriptConnect(connection_name, connection_ip, script_name, session_id_client, login, password)
    if rc == 'error':
        print("The connection \"{0}\" already exists".format(connection_name))

    connection = waitForConnection(icsClient, connection_name, connection_ip)
    if type(connection) == icsClient.connection:
        # success => activate data passthrough
        print("Activating Pass Through for \"{0}\".".format(script_name))
        connection.activateDataPassThrough()

    return connection

def waitForConnection(icsClient, connection_name, connection_ip):

    # this might take a while.
    print("Waiting for connection to be created ...")
    counter = 0
    connection = icsClient.findConnection(connection_name)
    while counter < 10 and type(connection) != icsClient.connection:
        time.sleep(1)
        connection = icsClient.findConnection(connection_name)
        counter = counter + 1

    if type(connection) != icsClient.connection:
        ## list is empty
        print("The connection \"{0}\" couldn't be created.".format(connection_name))
        return 0

    # this might take a while.
    print("Waiting for connection to {0} to be ready ...".format(connection_ip))
    counter = 0
    status = connection.getStatusConnectionStatus()
    while counter < 10 and status != "Configured" :
        time.sleep(1)
        new_status = connection.getStatusConnectionStatus()
        counter = counter + 1
        if new_status != status:
            status = new_status
            print("{0} sec : connection status == {1}".format(counter, status))

    if status != "Configured" :
        ## list is empty
        print("Connection to {0} couldn't be established.".format(connection_ip))
        return 0

    print("Connected to \"{0}\"".format(connection))
    return connection

def disconnectAll(icsClient):
    all_connections = icsClient.getConnections()
    for connection in all_connections:
        connection.disconnect()

def troubleshootConnection(connection):
    status = connection.getStatusConnectionStatus()
    if status != "Configured" :
        print("troubleshoot : The connection {0} isn't ready.".format(connection))
        return False
    print("troubleshoot : The connection {0} is ready.".format(connection))
    return True

def troubleshoot(icsClient):

    all_connections = icsClient.getConnections()
    if len(all_connections) == 0:
        print("troubleshoot : No connection defined.")
        return False

    for connection in all_connections:
        # check the connections first
        rc = troubleshootConnection(connection)
        # if rc == True:

    return True




import faulthandler
import json
import os
import queue
import sys
import threading
import time
import traceback
import types
from test import support

from test.libregrtest.runtest import (
    runtest, INTERRUPTED, CHILD_ERROR, PROGRESS_MIN_TIME,
    format_test_result)
from test.libregrtest.setup import setup_tests
from test.libregrtest.utils import format_duration


# Display the running tests if nothing happened last N seconds
PROGRESS_UPDATE = 30.0   # seconds

# If interrupted, display the wait progress every N seconds
WAIT_PROGRESS = 2.0   # seconds


def run_test_in_subprocess(testname, ns):
    """Run the given test in a subprocess with --slaveargs.

    ns is the option Namespace parsed from command-line arguments. regrtest
    is invoked in a subprocess with the --slaveargs argument; when the
    subprocess exits, its return code, stdout and stderr are returned as a
    3-tuple.
    """
    from subprocess import Popen, PIPE

    ns_dict = vars(ns)
    slaveargs = (ns_dict, testname)
    slaveargs = json.dumps(slaveargs)

    cmd = [sys.executable, *support.args_from_interpreter_flags(),
           '-u',    # Unbuffered stdout and stderr
           '-m', 'test.regrtest',
           '--slaveargs', slaveargs]
    if ns.pgo:
        cmd += ['--pgo']

    # Running the child from the same working directory as regrtest's original
    # invocation ensures that TEMPDIR for the child is the same when
    # sysconfig.is_python_build() is true. See issue 15300.
    popen = Popen(cmd,
                  stdout=PIPE, stderr=PIPE,
                  universal_newlines=True,
                  close_fds=(os.name != 'nt'),
                  cwd=support.SAVEDCWD)
    with popen:
        stdout, stderr = popen.communicate()
        retcode = popen.wait()
    return retcode, stdout, stderr


def run_tests_slave(slaveargs):
    ns_dict, testname = json.loads(slaveargs)
    ns = types.SimpleNamespace(**ns_dict)

    setup_tests(ns)

    try:
        result = runtest(ns, testname)
    except KeyboardInterrupt:
        result = INTERRUPTED, '', None
    except BaseException as e:
        traceback.print_exc()
        result = CHILD_ERROR, str(e)

    print()   # Force a newline (just in case)
    print(json.dumps(result), flush=True)
    sys.exit(0)


# We do not use a generator so multiple threads can call next().
class MultiprocessIterator:

    """A thread-safe iterator over tests for multiprocess mode."""

    def __init__(self, tests):
        self.interrupted = False
        self.lock = threading.Lock()
        self.tests = tests

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            if self.interrupted:
                raise StopIteration('tests interrupted')
            return next(self.tests)


class MultiprocessThread(threading.Thread):
    def __init__(self, pending, output, ns):
        super().__init__()
        self.pending = pending
        self.output = output
        self.ns = ns
        self.current_test = None
        self.start_time = None

    def _runtest(self):
        try:
            test = next(self.pending)
        except StopIteration:
            self.output.put((None, None, None, None))
            return True

        try:
            self.start_time = time.monotonic()
            self.current_test = test

            retcode, stdout, stderr = run_test_in_subprocess(test, self.ns)
        finally:
            self.current_test = None

        if retcode != 0:
            result = (CHILD_ERROR, "Exit code %s" % retcode, None)
            self.output.put((test, stdout.rstrip(), stderr.rstrip(),
                             result))
            return False

        stdout, _, result = stdout.strip().rpartition("\n")
        if not result:
            self.output.put((None, None, None, None))
            return True

        result = json.loads(result)
        assert len(result) == 3, f"Invalid result tuple: {result!r}"
        self.output.put((test, stdout.rstrip(), stderr.rstrip(),
                         result))
        return False

    def run(self):
        try:
            stop = False
            while not stop:
                stop = self._runtest()
        except BaseException:
            self.output.put((None, None, None, None))
            raise


def run_tests_multiprocess(regrtest):
    output = queue.Queue()
    pending = MultiprocessIterator(regrtest.tests)
    test_timeout = regrtest.ns.timeout
    use_timeout = (test_timeout is not None)

    workers = [MultiprocessThread(pending, output, regrtest.ns)
               for i in range(regrtest.ns.use_mp)]
    print("Run tests in parallel using %s child processes"
          % len(workers))
    for worker in workers:
        worker.start()

    def get_running(workers):
        running = []
        for worker in workers:
            current_test = worker.current_test
            if not current_test:
                continue
            dt = time.monotonic() - worker.start_time
            if dt >= PROGRESS_MIN_TIME:
                text = '%s (%s)' % (current_test, format_duration(dt))
                running.append(text)
        return running

    finished = 0
    test_index = 1
    get_timeout = max(PROGRESS_UPDATE, PROGRESS_MIN_TIME)
    try:
        while finished < regrtest.ns.use_mp:
            if use_timeout:
                faulthandler.dump_traceback_later(test_timeout, exit=True)

            try:
                item = output.get(timeout=get_timeout)
            except queue.Empty:
                running = get_running(workers)
                if running and not regrtest.ns.pgo:
                    print('running: %s' % ', '.join(running), flush=True)
                continue

            test, stdout, stderr, result = item
            if test is None:
                finished += 1
                continue
            regrtest.accumulate_result(test, result)

            # Display progress
            ok, test_time, xml_data = result
            text = format_test_result(test, ok)
            if (ok not in (CHILD_ERROR, INTERRUPTED)
                and test_time >= PROGRESS_MIN_TIME
                and not regrtest.ns.pgo):
                text += ' (%s)' % format_duration(test_time)
            elif ok == CHILD_ERROR:
                text = '%s (%s)' % (text, test_time)
            running = get_running(workers)
            if running and not regrtest.ns.pgo:
                text += ' -- running: %s' % ', '.join(running)
            regrtest.display_progress(test_index, text)

            # Copy stdout and stderr from the child process
            if stdout:
                print(stdout, flush=True)
            if stderr and not regrtest.ns.pgo:
                print(stderr, file=sys.stderr, flush=True)

            if result[0] == INTERRUPTED:
                raise KeyboardInterrupt
            test_index += 1
    except KeyboardInterrupt:
        regrtest.interrupted = True
        pending.interrupted = True
        print()
    finally:
        if use_timeout:
            faulthandler.cancel_dump_traceback_later()

    # If tests are interrupted, wait until tests complete
    wait_start = time.monotonic()
    while True:
        running = [worker.current_test for worker in workers]
        running = list(filter(bool, running))
        if not running:
            break

        dt = time.monotonic() - wait_start
        line = "Waiting for %s (%s tests)" % (', '.join(running), len(running))
        if dt >= WAIT_PROGRESS:
            line = "%s since %.0f sec" % (line, dt)
        print(line, flush=True)
        for worker in workers:
            worker.join(WAIT_PROGRESS)




## this python script can be used with icsClientPython_3
##      to test a connection to an scpi module connected at ttyS5



import sys
import time
import math
import argparse
import traceback
import importlib
import types
import os

import clientCommon

connectionName = "new_iCS_connection"
#connectionParameters = "ws://localhost:1234"
connectionLogin = "demo"
connectionPassword = "demo"

for argvIndex in range(0, len(sys.argv)):
    if argvIndex + 1 < len(sys.argv):
        if sys.argv[argvIndex] == "-c":
            connectionParameters = sys.argv[argvIndex + 1]
## this must specify the location of the icsClientPython_3 python module
pythonModulePath = "../platform/linux/64/"
if clientCommon.checkIsWindowsPlatform():
    pythonModulePath = "../bin/"
pythonModulePath = os.path.abspath(pythonModulePath)

## import icsClient module 
#icsClient = icsClientCommon.importFromPath("", "icsClientPython_for_iCS_3")	#Library from last year
icsClient = clientCommon.importFromPath(pythonModulePath, "icsClientPython_3")	#Windows
#icsClient = icsClientCommon.importFromPath("/home/jr/iseg/isegcontrol_DEV_191021_64/platform/linux/64a/", "libicsClientPython3") #Ubuntu Linux 18.04
if icsClient == 0:
    print("Failed to import icsClient module => Exit.")
    exit(1)

## establish a connection and wait until it is ready (with timeout)
connection = clientCommon.establishConnection(icsClient, connectionName, connectionParameters, connectionLogin, connectionPassword)
if type(connection) != icsClient.connection:
    print("Failed to connect => Exit.")
    exit(1)

##


power_frame = "{\"i\":\"<session_id>\",\"t\":\"request\",\"c\":[{\"c\":\"getItem\",\"p\":{\"p\":{\"l\":\"0\",\"a\":\"1000\",\"c\":\"\"},\"i\":\"Control.power\",\"v\":\"\",\"u\":\"\"}}],\"r\":\"websocket\"}"

connection.rawCommand(power_frame)

print(power_frame)

result = connection.rawCommand("{\"i\":\"<session_id>\",\"t\":\"request\",\"c\":[{\"c\":\"setItem\",\"p\":{\"p\":{\"l\":\"0\",\"a\":\"1000\",\"c\":\"\"},\"i\":\"Control.power\",\"v\":\"1\",\"u\":\"\"}}],\"r\":\"websocket\"}")
print(result)

all_crates = connection.getCrates()
print("All crates {0}.".format(all_crates))

all_modules = connection.getModules()
print("All modules {0}.".format(all_modules))

all_channels = all_modules[0].channels
print("All channels {0}.".format(all_channels))
time.sleep(5)

def outputModule(item, function):
    time.sleep(0.01)
    print(item  + str(function))

def outputModuleInt(item, function):
    time.sleep(0.01)
    print(item  + str(int(function)))

outputModule("Firmwarename: ", all_modules[0].getStatusFirmwareName())
outputModuleInt("Seriennummer: ", all_modules[0].getStatusSerialNumber())
outputModule("Release: ", all_modules[0].getStatusFirmwareRelease())
outputModule("Temperatur: ", all_modules[0].getStatusTemperature())

#all_channels[1].setControlVoltageSet("123")

def output(channel, item, function):
    time.sleep(0.01)
    print(item + str(channel) + ": " + str(function))
    

def outputChannelItems():
    all_channels = all_modules[0].channels
    for channel in range(0, len(all_channels)):
        output(channel, "Status.voltageNominal", all_channels[channel].getStatusVoltageNominal())
    for channel in range(0, len(all_channels)):
        output(channel, "Control.voltageSet", all_channels[channel].getControlVoltageSet())
    for channel in range(0, len(all_channels)):
        output(channel, "Status.voltageMeasure", all_channels[channel].getStatusVoltageMeasure()) 
    for channel in range(0, len(all_channels)):
        output(channel, "Status.currentNominal", all_channels[channel].getStatusCurrentNominal())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Control.voltageBounds", all_channels[channel].getControlVoltageBounds())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Control.voltageRampspeedDown", all_channels[channel].getControlVoltageRampspeedDown())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Control.voltageRampspeedUp", all_channels[channel].getControlVoltageRampspeedUp())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Status.voltageRampspeedMin", all_channels[channel].getStatusVoltageRampspeedMin())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Status.voltageRampspeedMax", all_channels[channel].getStatusVoltageRampspeedMax())
#    for channel in range(0, len(all_channels)):
#       output(channel, "Control.voltageVctCoefficient", all_channels[channel].getControlVctCoefficient())
    for channel in range(0, len(all_channels)):
        output(channel, "Control.currentSet", all_channels[channel].getControlCurrentSet())
    for channel in range(0, len(all_channels)):
        output(channel, "Status.currentMeasure", all_channels[channel].getStatusCurrentMeasure())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Control.currentBounds", all_channels[channel].getControlCurrentBounds())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Control.currentRampspeedDown", all_channels[channel].getControlCurrentRampspeedDown())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Control.currentRampspeedUp", all_channels[channel].getControlCurrentRampspeedUp())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Status.currentRampSpeedMin", all_channels[channel].getStatusCurrentRampspeedMin())
#    for channel in range(0, len(all_channels)):
#        output(channel, "Status.currentRampSpeedMax", all_channels[channel].getStatusCurrentRampspeedMax())
    for channel in range(0, len(all_channels)):
        output(channel, "Status.voltageLimitExceeded", all_channels[channel].getStatusVoltageLimitExceeded())
    for channel in range(0, len(all_channels)):
        output(channel, "Status.currentLimitExceeded", all_channels[channel].getStatusCurrentLimitExceeded())
    for channel in range(0, len(all_channels)):
        output(channel, "Status.currentTrip", all_channels[channel].getStatusTrip())
    for channel in range(0, len(all_channels)):
        output(channel, "Status.on", all_channels[channel].getStatusOn())
    for channel in range(0, len(all_channels)):
        output(channel, "Status.emergenc", all_channels[channel].getStatusEmergency())
    for channel in range(0, len(all_channels)):
        output(channel, "Setup.delayedTripAction", all_channels[channel].getSetupDelayedTripAction())
    for channel in range(0, len(all_channels)):
        output(channel, "Setup.delayedTripTime", all_channels[channel].getSetupDelayedTripTime())

outputChannelItems()


#sys.stdout = open("ics-python3-channel-items.txt", "w")
#print(help(all_channels[0]))
#sys.stdout.close()

#sys.stdout = open("ics-python3-module-items.txt", "w")
#print(help(all_modules[0]))
#sys.stdout.close()

#sys.stdout = open("ics-python3-crates-items.txt", "w")
#print(help(all_crates[0]))
#sys.stdout.close()

clientCommon.disconnectAll(icsClient)

